package org.processmining.security.plugins;

import java.io.File;
import java.util.Collection;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.in.XMxmlGZIPParser;
import org.deckfour.xes.in.XesXmlGZIPParser;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XLog;
import org.processmining.connection.PAlignmentResultConnection;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.log.csv.CSVFile;
import org.processmining.log.csv.CSVFileReferenceOpenCSVImpl;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.partialorder.plugins.PartialNetReplayer;
import org.processmining.partialorder.ptrace.model.PLog;
import org.processmining.petrinets.list.impl.PetriNetListImpl;
import org.processmining.petrinets.list.plugin.ImportPetriNetListPlugin;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.security.plugins.criterion.CriterionFunction;
import org.processmining.security.plugins.interDataProcessPolicy.CRUD;
import org.processmining.security.plugins.replayer.InterLevelReplayResult;
import org.processmining.security.plugins.replayer.InterLevelReplayerParameters;
import org.processmining.security.plugins.replayer.InterLevelResultConnection;
import org.processmining.security.plugins.replayer.MultiThreadInterLevelReplayer;
import org.processmining.security.plugins.replayer.aStar.CostFunction;
import org.processmining.security.plugins.ui.CostFunctionPanel;
import org.processmining.security.plugins.ui.CriterionConfigurationPanel;

@Plugin(name = "Inter-Level Replayer", returnLabels = { "Inter-Level Replayer Results" }, returnTypes = { InterLevelReplayResult.class },
		parameterLabels = { "Petri net", "Process Event Log", "System Event Log", "CRUD", "Partial Order Replayer Result" }, categories = { PluginCategory.ConformanceChecking },
		keywords = {}, help = "Compute mappings between traces", userAccessible = true)
public class InterLevelConformanceCheckingPlugin {
	public static boolean UI = true;
	public static boolean FID_AVAILABLE = true;
	public static String prefixAddr = "./example/fineManagementProcess/";
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "", email = "",
			uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "", requiredParameterLabels = { 0, 1, 2, 3})
	public InterLevelReplayResult replay(UIPluginContext context, Petrinet net, XLog processEventLog, XLog systemEventLog, CSVFile crudCSVFile) throws Exception {
		
		context.log("Alignment Start");
		PNRepResult alignments = new PartialNetReplayer().replayLogGUIStandard(context, net, processEventLog);
		context.log("Alignment End");		
		
		InterLevelReplayerParameters params = getReplayerParameters(context, alignments, systemEventLog);
		if(params == null)
			throw new Exception("The required parameters are not set");

		return replay(context, alignments, systemEventLog, crudCSVFile, params);
	}
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "", email = "",
			uiLabel = UITopiaVariant.USEVARIANT, pack = "")
	@PluginVariant(variantLabel = "", requiredParameterLabels = { 4, 2, 3})
	public InterLevelReplayResult replay(UIPluginContext context, PNRepResult alignments, XLog systemEventLog, CSVFile crudCSVFile) throws Exception {
		InterLevelReplayerParameters params = getReplayerParameters(context, alignments, systemEventLog);
		if(params == null)
			throw new Exception("The required parameters are not set");
		return replay(context, alignments, systemEventLog, crudCSVFile, params);
	}
	private InterLevelReplayResult replay(UIPluginContext context, PNRepResult alignments, XLog systemEventLog, CSVFile crudCSVFile, InterLevelReplayerParameters params) throws Exception{
		CRUD crud = new CRUD(crudCSVFile);
		
		context.log("Inter Data Process Replayer Start");
		InterLevelReplayResult res = new MultiThreadInterLevelReplayer().perform(context, alignments, systemEventLog, crud, params);
		context.log("Inter Data Process Replayer End");
		
		context.addConnection(new InterLevelResultConnection("RepResults", res, alignments, crud, systemEventLog));
		return res;
	}
	
	public InterLevelReplayerParameters getReplayerParameters(UIPluginContext context, PNRepResult alignments, XLog systemEventLog) throws ConnectionCannotBeObtained{
		PAlignmentResultConnection conn = context.getConnectionManager().getFirstConnection(PAlignmentResultConnection.class, context, alignments);
		PLog poLog = conn.getObjectWithRole(PAlignmentResultConnection.PARTIALLOG);
		XLogInfo processEventLogInfo = XLogInfoFactory.createLogInfo(poLog.getXLog()); 
		XLogInfo systemEventLogInfo = XLogInfoFactory.createLogInfo(systemEventLog);
		Collection<String> pAttributes = processEventLogInfo.getEventAttributeInfo().getAttributeKeys();
		Collection<String> sAttributes = systemEventLogInfo.getEventAttributeInfo().getAttributeKeys();
		
		CostFunctionPanel costFunctionPanel = new CostFunctionPanel();
		InteractionResult result = null;
		if(UI){
			result = context.showConfiguration("Cost Function", costFunctionPanel);
			if (result == InteractionResult.CANCEL) {
				context.getFutureResult(0).cancel(true);
				return null;
			}		
		}
		CostFunction costFunction = new CostFunction(costFunctionPanel.getCosts());
		
		CriterionConfigurationPanel panel = new CriterionConfigurationPanel(pAttributes, sAttributes);			
		if(UI){
			result = context.showConfiguration("Criterion Function", panel);
			if (result == InteractionResult.CANCEL) {
				context.getFutureResult(0).cancel(true);
				return null;
			}		
		}
		CriterionFunction criterionFunction = panel.getCriterion();
		InterLevelReplayerParameters parameters = new InterLevelReplayerParameters(costFunction, criterionFunction);
		return parameters;
	}
	
//	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "", email = "",
//			uiLabel = UITopiaVariant.USEVARIANT, pack = "")
//	@PluginVariant(variantLabel = "", requiredParameterLabels = { })
	public InterLevelReplayResult replayWithoutParams(UIPluginContext context) throws Exception {
		XLog sysEventLog = loadEventLog(prefixAddr + "SystemEventLog.xes.gz");
		XLog procEventLog = loadEventLog(prefixAddr + "ProcessEventLog.xes.gz");
		File csv = new File(prefixAddr+"CRUD.csv");
		CSVFile csvFile = new CSVFileReferenceOpenCSVImpl(csv.toPath());
		File netFile = new File(prefixAddr + "ProcessModel.pnml");
		Petrinet petriNet = ((PetriNetListImpl) new ImportPetriNetListPlugin().importFile(context, netFile)).get(0);
		
		return replay(context, petriNet, procEventLog, sysEventLog, csvFile);
	}
	
	public static XLog loadEventLog(String addr){
		 File eventLog = new File(addr);		
		 XesXmlGZIPParser parser = new XesXmlGZIPParser();		 
		 try {
			 if(parser.canParse(eventLog)){
				 	return parser.parse(eventLog).get(0);
			 }else  if(new XMxmlGZIPParser().canParse(eventLog))
				 return new XMxmlGZIPParser().parse(eventLog).get(0);
			 else
				 System.err.println("Cannot Parse Traces!!!!");
		 } catch (Exception e) {
			 e.printStackTrace();
		 }
		return null;
		 
	}
	

}