package org.processmining.security.plugins.replayer;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.PriorityQueue;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.partialorder.models.palignment.Move;
import org.processmining.partialorder.models.palignment.PAlignment;
import org.processmining.security.plugins.Utility;
import org.processmining.security.plugins.criterion.CriterionFunction;
import org.processmining.security.plugins.heuristic.HeuristicFunction;
import org.processmining.security.plugins.heuristic.HeuristicFunctionImp;
import org.processmining.security.plugins.interDataProcessPolicy.CRUD;
import org.processmining.security.plugins.interDataProcessPolicy.CRUDEntry;
import org.processmining.security.plugins.replayer.aStar.AStartNode;
import org.processmining.security.plugins.replayer.aStar.CostFunction;


public class InterLevelReplayer extends Thread{
	
	
	private InterLevelReplayResult replayResult;
	private CRUD crud;
	private CostFunction cost;
	private CriterionFunction criterionFunction;
	private PAlignment pAlignment;
	private XTrace systemTrace;
	private InterLevelReplayResult list;		
	private UIPluginContext context;
	public static int counter;
	public InterLevelReplayer(UIPluginContext context, InterLevelReplayResult list, PAlignment pAlignment, XTrace systemTrace, 
			CRUD crud, InterLevelReplayerParameters params) throws Exception{
		this.context = context;
		this.crud = crud;
		this.criterionFunction = params.getCriterionFunction();
		this.cost = params.getCostFunction();
		this.pAlignment = pAlignment;
		this.systemTrace = systemTrace;
		this.list = list;		
	}
	public void run()
	{
		InterLevelSyncReplayResult interLevelAlignment = null;
		try {
			interLevelAlignment = findOptimalInterLevelAlignment(pAlignment, systemTrace);
			if(interLevelAlignment == null)
				System.err.println("The constructed inter-level alignment is null! Case ID: " + ((XAttributeLiteralImpl) systemTrace.getAttributes().get("concept:name")).getValue());			
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		interLevelAlignment.setPAlignment(pAlignment);
		list.add(interLevelAlignment);	
		counter++;
		if(counter % 10000 == 0)
			context.log(counter + " inter-level alignments have been constructed");
	}
	private InterLevelSyncReplayResult findOptimalInterLevelAlignment(PAlignment pAlignment, XTrace systemTrace) throws Exception{
		HeuristicFunction heuristicFunction = new HeuristicFunctionImp(pAlignment, systemTrace, cost, crud, criterionFunction);	
		PriorityQueue<AStartNode> pQueue = new PriorityQueue<AStartNode>();
		pQueue.add(init(pAlignment, systemTrace));			
		while(pQueue.size() != 0){			
			AStartNode node = pQueue.poll();
			if(isTargetNode(pAlignment, node)){
				return node.getSyncReplayResult();
			}
			List<AStartNode> newNodes = successor(pAlignment, systemTrace, node);
			for (AStartNode aStartNode : newNodes) {
				int lastIndex = aStartNode.getSyncReplayResult().getCompositeMoves().size() - 1;
				CompositeMove lastCM = aStartNode.getSyncReplayResult().getCompositeMoves().get(lastIndex);
				double g_cost = aStartNode.getF_cost() - aStartNode.getH_cost() + 
						cost.getCompositeMoveCost(lastCM.getIDPStepType(), lastCM.getAlignmentMove());
				double h_cost = heuristicFunction.getDistanceToTarget(systemTrace, aStartNode.getNextSystemEventIndex());
				double f_cost = g_cost + h_cost;
				aStartNode.setCost(f_cost, h_cost);
			}
			pQueue.addAll(newNodes); 
		}
		
		return null;
	}
	
	public AStartNode[] removeNodesWithMaxCost(PriorityQueue<AStartNode> pQueue, int size){
		AStartNode nodesWithMaxCost[] = new AStartNode[size];
		int index = 0;
		int minCostIndex = 0;
		for (AStartNode node : pQueue) {			
			if(index < size){
				nodesWithMaxCost[index] = node;
				if(node.getF_cost() < nodesWithMaxCost[minCostIndex].getF_cost())
					minCostIndex = index;
				index++;
			}else {
				if(node.getF_cost() > nodesWithMaxCost[minCostIndex].getF_cost())
				{
					nodesWithMaxCost[minCostIndex] = node;
					for(int i = 0 ; i < size; i++)
						if(nodesWithMaxCost[i].getF_cost() < nodesWithMaxCost[minCostIndex].getF_cost())
							minCostIndex = i;
				}
			}
		}
		return nodesWithMaxCost;
	}
	private AStartNode init(PAlignment pAlignment, XTrace systemTrace){
		List<Move> initialMoves = new ArrayList<Move>();
		List<Move> nextMoves = new ArrayList<Move>();
		HashMap<Move, List<CRUDEntry>> remainingActions = new HashMap<Move, List<CRUDEntry>>();
		for (Move move : pAlignment.getMoves()) { // Find moves with no parent
			if(pAlignment.getDirectParents(move).size() == 0){
				initialMoves.add(move);
				remainingActions.put(move, crud.getRequiredActions(move));				
			}
		}
		String caseID = ((XAttributeLiteralImpl) systemTrace.getAttributes().get("concept:name")).getValue();
		AStartNode initNode = new AStartNode(new ArrayList<Move>(), initialMoves, nextMoves, systemTrace.size() > 0 ? 0 : -1, 
				new InterLevelSyncReplayResult(caseID), remainingActions, 0, 0);
		for (Move move : initialMoves) {//Initialize next moves
			for(Move nMove : initNode.filterChildrenWithObservedParents(pAlignment, pAlignment.getDirectChildren(move))){
				if(!nextMoves.contains(nMove)){
					nextMoves.add(nMove);
					remainingActions.put(nMove, crud.getRequiredActions(nMove));
				}
			}
		}
		initNode.setNext_moves(nextMoves);
		return initNode;
	}
	private boolean isTargetNode(PAlignment pAlignment, AStartNode node){
		if(node.getNextSystemEventIndex() != -1)
			return false;
		if(node.getNext_moves().size() > 0)
			return false;
		for (Move move : node.getCurr_moves()) {
			List<CRUDEntry> crudEntry = crud.getRequiredActions(Utility.getActivity(move));
			if(!node.hasAllRequiredActions(crudEntry))
				return false;
		}
		return true;
	}
	
	private List<AStartNode> successor(PAlignment pAlignment, XTrace systemTrace, AStartNode node) throws Exception {
		List<AStartNode> nextNodes = new ArrayList<AStartNode>();
		CompositeMove compositeMove = null;
		if(node.getNextSystemEventIndex() != -1){
			//No Context					
			XEvent systemEvent = systemTrace.get(node.getNextSystemEventIndex());
			compositeMove = new CompositeMove(systemEvent, null, null);
			if(criterionFunction.isValid(pAlignment,node.getSyncReplayResult(), compositeMove))
				addNewMove(nextNodes, node, compositeMove, getNextSystemEvent(node.getNextSystemEventIndex(), systemTrace));
			//Alignment + SystemEvent
			for (Move move: node.getCurr_moves()) {				
				CRUDEntry crudEntry = crud.crudCheck(Utility.getActivity(move), systemTrace.get(node.getNextSystemEventIndex()));
				if(crudEntry != null && node.contain(move, crudEntry))
					continue;
				if(crud.getRequiredActions(move).size() == 0 && node.hasMoveWithNoAction(move))
					continue;
				//prune the search space
				if(node.hasMissingAction(move))
					continue;
				compositeMove = new CompositeMove(systemEvent, move, crudEntry);
				if(criterionFunction.isValid(pAlignment,node.getSyncReplayResult(), compositeMove))					
					addNewMove(nextNodes, node, compositeMove, getNextSystemEvent(node.getNextSystemEventIndex(), systemTrace));
			} // Check nextMoves
	//		List<CompositeMove> cmList = new ArrayList<CompositeMove>();
			for (Move move : node.getNext_moves()) {
				if(node.checkParentsActions(pAlignment.getDirectParents(move))){
					CRUDEntry crudEntry = crud.crudCheck(Utility.getActivity(move), systemTrace.get(node.getNextSystemEventIndex()));
					CompositeMove cm = new CompositeMove(systemEvent, move, crudEntry);
					if(criterionFunction.isValid(pAlignment,node.getSyncReplayResult(), cm))				
						addNewMove(nextNodes, node, cm, getNextSystemEvent(node.getNextSystemEventIndex(), systemTrace), pAlignment);					
				}
			}
		}
		//Alignment + Missing Action/NoAction
		for (Move move: node.getCurr_moves()) {
			if(crud.getRequiredActions(move).size() == 0 && !node.contain(move)){
				CompositeMove compositeMove1 = new CompositeMove(null, move, null);
				if(criterionFunction.isValid(pAlignment,node.getSyncReplayResult(), compositeMove1))
					addNewMove(nextNodes, node, compositeMove1, node.getNextSystemEventIndex());				
			}else {
				for (CRUDEntry crudEntry : node.getRemainingActions(move)) {
					if(isOrderOfMissingActionsCorrect(node, move, crudEntry)){
						compositeMove = new CompositeMove(null, move, crudEntry);
						if(criterionFunction.isValid(pAlignment,node.getSyncReplayResult(), compositeMove))
							addNewMove(nextNodes, node, compositeMove, node.getNextSystemEventIndex());
					}
				}		
			}
		}// Check nextMoves		
		for (Move move : node.getNext_moves()) {
			//Activity without any required action
			if(node.checkParentsActions(pAlignment.getDirectParents(move))){
				if(crud.getRequiredActions(move).size() == 0 && !node.contain(move)){
					compositeMove = new CompositeMove(null, move, null);
					if(criterionFunction.isValid(pAlignment,node.getSyncReplayResult(), compositeMove))	
						addNewMove(nextNodes, node, compositeMove, node.getNextSystemEventIndex(), pAlignment);
				}else {
					for (CRUDEntry crudEntry : node.getRemainingActions(move)) {
						if(isOrderOfMissingActionsCorrect(node, move, crudEntry)){
							compositeMove = new CompositeMove(null, move, crudEntry);
							if(criterionFunction.isValid(pAlignment,node.getSyncReplayResult(), compositeMove))
								addNewMove(nextNodes, node, compositeMove, node.getNextSystemEventIndex(), pAlignment);
						}
					}	
				}
			}
		}
		return nextNodes;
	}
	private boolean isOrderOfMissingActionsCorrect(AStartNode node, Move move, CRUDEntry crudEntry){
		for (CompositeMove cm : node.getSyncReplayResult().getCompositeMoves()) 
			if(cm.getAlignmentMove() != null && move.equals(cm.getAlignmentMove()) && cm.getSystemEvent() == null && 
			cm.getCRUDEntry() != null && cm.getCRUDEntry().toString().compareTo(crudEntry.toString()) <= 0)
				return false;
		return true;
	}
	private void addNewMove(List<AStartNode> nextNodes, AStartNode node, CompositeMove cm, int index) throws CloneNotSupportedException{
		AStartNode newNode = (AStartNode) node.clone();
		newNode.addCompositeMove(cm, index);
		nextNodes.add(newNode);
	}
	private void addNewMove(List<AStartNode> nextNodes, AStartNode node, CompositeMove cm, int index, PAlignment pAlignment) throws CloneNotSupportedException{
		AStartNode newNode = (AStartNode) node.clone();
		newNode.addCompositeMove(cm, index, true, pAlignment, crud);
		nextNodes.add(newNode);
	}
	private int getNextSystemEvent(int index, XTrace systemTrace){
		if(index < systemTrace.size() - 1)
			return index + 1;
		else
			return -1;
	}
	public InterLevelReplayResult getReplayResult() {
		return replayResult;
	}	
	

}
