package org.processmining.security.plugins.replayer.aStar;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;

import org.processmining.partialorder.models.palignment.Move;
import org.processmining.partialorder.models.palignment.PAlignment;
import org.processmining.security.plugins.interDataProcessPolicy.CRUD;
import org.processmining.security.plugins.interDataProcessPolicy.CRUDEntry;
import org.processmining.security.plugins.replayer.CompositeMove;
import org.processmining.security.plugins.replayer.InterLevelSyncReplayResult;

public class AStartNode implements Cloneable, Comparable<AStartNode>{
	int nextSystemEventIndex;
	InterLevelSyncReplayResult syncReplayResult;
	List<Move> past_moves, curr_moves, next_moves;
	double f_cost, h_cost;
	HashMap<Move, List<CRUDEntry>> remainingActions;
	public AStartNode(List<Move> used_moves, List<Move> curr_moves, List<Move> next_moves, int systemEvent, 
			InterLevelSyncReplayResult syncReplayResult, HashMap<Move, List<CRUDEntry>> map, double f_cost, double h_cost){
		this.nextSystemEventIndex = systemEvent;
		this.past_moves = used_moves;
		this.curr_moves = curr_moves;
		this.next_moves = next_moves;
		this.syncReplayResult = syncReplayResult;
		this.f_cost = f_cost;
		this.h_cost = h_cost;
		remainingActions = map;
	}
	public InterLevelSyncReplayResult getSyncReplayResult() {
		return syncReplayResult;
	}
	public void setCost(double f_cost, double h_cost){
		this.f_cost = f_cost;
		this.h_cost = h_cost;
	}
	public double getF_cost() {
		return f_cost;
	}
	public double getH_cost() {
		return h_cost;
	}
	public int getNextSystemEventIndex() {
		return nextSystemEventIndex;
	}
	public List<CRUDEntry> getRemainingActions(Move move) {
		return remainingActions.get(move);
	}
	public void addCompositeMove(CompositeMove cm, int nextSystemEventIndex, boolean isNextMove, PAlignment pAlignment, CRUD crud){
		addCompositeMove(cm, nextSystemEventIndex);
		if(isNextMove){
			Collection<Move> parentNodes = pAlignment.getDirectParents(cm.getAlignmentMove());
			next_moves.remove(cm.getAlignmentMove());
			curr_moves.add(cm.getAlignmentMove());			
			for (Move move : parentNodes) {
				past_moves.add(move);
				curr_moves.remove(move);
				remainingActions.remove(move);
			}
			for(Move move : filterChildrenWithObservedParents(pAlignment, pAlignment.getDirectChildren(cm.getAlignmentMove()))){
				next_moves.add(move);
				remainingActions.put(move, crud.getRequiredActions(move));
			}
		}
	}
	public void addCompositeMove(CompositeMove cm, int nextSystemEventIndex){
		this.nextSystemEventIndex = nextSystemEventIndex;
		syncReplayResult.addCompositeMove(cm);
		
		if(cm.getAlignmentMove() != null && cm.getCRUDEntry() != null && remainingActions.get(cm.getAlignmentMove()) != null)
			remainingActions.get(cm.getAlignmentMove()).remove(cm.getCRUDEntry());				
	}
	
	public boolean checkParentsActions(Collection<Move> moves){
		for (Move move : moves) {
			if(!contain(move))
				return false;
			if(!curr_moves.contains(move) && !past_moves.contains(move))
				return false;
			if(remainingActions.get(move) != null && remainingActions.get(move).size() > 0)
				return false;
		}
		return true;
	}
	public List<Move> getCurr_moves() {
		return curr_moves;
	}
	public List<Move> getNext_moves() {
		return next_moves;
	}
	public boolean hasAllRequiredActions(List<CRUDEntry> crudEntry){
		for (CRUDEntry entry : crudEntry) {
			boolean flag = false;
			for (CompositeMove compositeMove : syncReplayResult.getCompositeMoves()) {				
				if(compositeMove.getCRUDEntry() != null && compositeMove.getCRUDEntry().compareTo(entry) == 0){
					flag = true;
					break;
				}
			}
			if(!flag)
				return false;
		}
		return true;
	}
	public boolean contain(Move move){
		for (CompositeMove cm : syncReplayResult.getCompositeMoves()) {
			if(cm.getAlignmentMove() != null && cm.getAlignmentMove().equals(move))
				return true;
		}
		return false;
	}
	public boolean hasMoveWithNoAction(Move move){
		for (CompositeMove cm : syncReplayResult.getCompositeMoves()) {
			if(cm.getAlignmentMove() != null && cm.getCRUDEntry() == null && 
					cm.getSystemEvent() == null && cm.getAlignmentMove().equals(move))
				return true;
		}
		return false;		
	}
	public boolean contain(Move move, CRUDEntry entry){
		for (CompositeMove cm : syncReplayResult.getCompositeMoves()) {
			if(cm.getAlignmentMove() != null && cm.getCRUDEntry() != null && 
					cm.getCRUDEntry().equals(entry) && cm.getSystemEvent() == null && cm.getAlignmentMove().equals(move))
				return true;
		}
		return false;
	}
	public boolean hasMissingAction(Move move){
		for (CompositeMove cm : syncReplayResult.getCompositeMoves()) {
			if(cm.getAlignmentMove() != null && cm.getAlignmentMove().equals(move) && cm.getCRUDEntry() != null && cm.getSystemEvent() == null)
				return true;
		}
		return false;
	}
	public AStartNode clone() throws CloneNotSupportedException {
		HashMap<Move, List<CRUDEntry>> map = new HashMap<Move, List<CRUDEntry>>();
		for (Move move : remainingActions.keySet()) {
			map.put(move,  new ArrayList<CRUDEntry>(remainingActions.get(move)));
		}
		
		return new AStartNode(new ArrayList<Move>(past_moves), 
				new ArrayList<Move>(curr_moves), new ArrayList<Move>(next_moves), nextSystemEventIndex, syncReplayResult.clone(), 
				map, f_cost, h_cost);
	}
	public Collection<Move> filterChildrenWithObservedParents(PAlignment pAlignment, Collection<Move> children){
		Collection<Move> validMoves = new ArrayList<Move>();
		for(Move child : children){
			boolean flag = true;
			for (Move parent : pAlignment.getDirectParents(child)) {
				if(!curr_moves.contains(parent) && !past_moves.contains(parent)){
					flag = false;
					break;
				}
			}
			if(flag)
				validMoves.add(child);
		}
		return validMoves;
	}
	public void setNext_moves(List<Move> next_moves) {
		this.next_moves = next_moves;
	}
	public String toString() {
		return syncReplayResult.toString();
	}
	public int compareTo(AStartNode o) {
		return (int) Math.ceil(getF_cost() - o.getF_cost());
	}
}
