/**
 * 
 */
package org.processmining.security.plugins.visualization.sequence;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.font.TextAttribute;
import java.text.AttributedString;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.Date;

import javax.swing.JComponent;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.extension.std.XExtendedEvent;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.model.XTrace;
import org.processmining.partialorder.models.palignment.Move;
import org.processmining.plugins.petrinet.visualization.AlignmentConstants;
import org.processmining.security.plugins.Utility;
import org.processmining.security.plugins.replayer.CompositeMove;
import org.processmining.security.plugins.replayer.CompositeMoveType;
import org.processmining.security.plugins.replayer.InterLevelSyncReplayResult;

/**
 * @author aadrians
 * This class is modification of class ProcessInstanceView
 * package org.processmining.plugins.log.ui.logdialog
 * made by Christian W. Guenther (christian@deckfour.org)
 * 
 * The original class need to be re-implemented the core visualization is different
 *  
 */
public class InterLevelInstanceConformanceView extends JComponent implements MouseListener, MouseMotionListener {
	
	private static final long serialVersionUID = -3966399549433794592L;
	protected static Color colorAttenuationDark = new Color(0, 0, 0, 160);
	protected static Color colorAttenuationBright = new Color(0, 0, 0, 80);
	protected static Color colorBgInstanceflag = new Color(70, 70, 70, 210);
	protected static Color colorBgEventFlag = new Color(30, 30, 30, 200);  
	
	protected static Color colorModelMove = new Color(75,204,255);
	
	protected static DecimalFormat format = new DecimalFormat("##0.00%");
	protected static DateFormat dateFormat = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss.SSS");

	protected static int trackPadding = 120;
	protected static int trackY_highLevel = 20;
	protected static int trackY_lowLevel = 80;
	protected static int trackHeight = 35;
	protected final int elementWidth; 
	protected final int highLevelEventWidth;
	protected static int elementTriOffset = 6;
	protected static int spaceBetweenMoves = 19;
	protected static int spaceBetweenSystemEvents = 33;
//	protected static int spaceBetweenMoves = 25;
//	protected static int spaceBetweenSystemEvents = 7;
	
	protected XLogInfo info;
	protected String traceLabel;
	protected int maxOccurrenceCount;
	protected XTrace instance;
	protected boolean mouseOver = false, isHighLevelStepSelected;
	protected int mouseX;
	protected int mouseY;
	protected Stroke lineStroke, dashed, boarder;
	protected InterLevelSyncReplayResult replayResult;
	private int numberOfActions;
	private int eventNum;
	protected InterLevelReplayResultVisualizationPanel visualizationPanel;

	public InterLevelInstanceConformanceView(InterLevelReplayResultVisualizationPanel visualizationPanel, String traceLabel, InterLevelSyncReplayResult replayResult, int elementWidth, int highLevelEventWidth) {
		this.visualizationPanel = visualizationPanel;
		this.elementWidth = elementWidth;
		this.highLevelEventWidth = highLevelEventWidth;
		this.traceLabel = traceLabel;
		this.replayResult = replayResult;
		addMouseListener(this);
		addMouseMotionListener(this);
		
		numberOfActions = 0;
		for (CompositeMove cm : replayResult.getCompositeMoves()) {
			if(cm.getSystemEvent() != null || cm.getCRUDEntry() != null)
				numberOfActions++;
		}
		
		int width = (numberOfActions * elementWidth) + trackPadding + 300;
		setMinimumSize(new Dimension(width, 120));
		setMaximumSize(new Dimension(Integer.MAX_VALUE, 115));
		setPreferredSize(new Dimension(width, 120));
		setDoubleBuffered(true);
		lineStroke = new BasicStroke(2.0f);
		dashed = new BasicStroke(2.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 0, new float[]{5, 5}, 0);
		boarder = new BasicStroke();		
	}
	
	@Override
	protected void paintComponent(Graphics g) {
		Rectangle clip = getVisibleRect();//g.getClipBounds();
		Graphics2D g2d = (Graphics2D) g;
		g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		// draw background
		g2d.setColor(Color.white);//new Color(30, 30, 30));
		g2d.fillRect(clip.x, clip.y, clip.width, clip.height);
		// determine active event
		int activeHLEvent = -1, activeLLEvent = -1;
		if (mouseOver == true) {
			activeLLEvent = mapLowLevelEventIndex(mouseX, mouseY);
			activeHLEvent = mapHighLevelEventIndex(mouseX, mouseY);
		}
		// draw events
		int clipRightX = clip.x + clip.width;
		int trackRightX;
		if (replayResult != null){
			trackRightX = trackPadding + (numberOfActions * (elementWidth + spaceBetweenSystemEvents));
		} else {
			trackRightX = trackPadding + (instance.size() * (elementWidth + spaceBetweenSystemEvents));
		}
		int startX = clip.x - (clip.x % (elementWidth + spaceBetweenSystemEvents)); // shift to left if necessary
		if (startX < trackPadding) {
			startX = trackPadding;
		}
		
		int startX_HL = clip.x - (clip.x % (highLevelEventWidth + spaceBetweenMoves)); // shift to left if necessary
		if (startX_HL < trackPadding) {
			startX_HL = trackPadding;
		}
		/*High Level alignment*/
		int eventIndex_ = (startX_HL - trackPadding) / (highLevelEventWidth + spaceBetweenMoves);
		int numOfSteps = replayResult.getSeqAlignment().size();
		int trackRightX_ = trackPadding + ( numOfSteps * (highLevelEventWidth + spaceBetweenMoves));
		for (int x = startX_HL; ((x < clipRightX) && (x < trackRightX_)); x += highLevelEventWidth + spaceBetweenMoves) {
			drawEvent(g2d, eventIndex_, (eventIndex_ == activeHLEvent), x , trackY_highLevel, highLevelEventWidth, trackHeight, true);
			eventIndex_++;
		}
		//***Low level***
		eventNum = 0;
		int eventIndex = (startX - trackPadding) / (elementWidth + spaceBetweenSystemEvents);
		for (int x = startX; ((x < clipRightX) && (x < trackRightX)); x += elementWidth + spaceBetweenSystemEvents) {
			boolean isAssociatedMoveSelected = false;
			if(activeHLEvent != -1){
				Move move = replayResult.getCompositeMovesWithAction().get(eventIndex).getAlignmentMove();
				Move selectedMove = replayResult.getSeqAlignment().get(activeHLEvent);
				 if(move != null && selectedMove.equals(move))
					isAssociatedMoveSelected = true;
			}
			drawEvent(g2d, eventIndex, (eventIndex == activeLLEvent) || isAssociatedMoveSelected, x, trackY_lowLevel, elementWidth , elementWidth, false);
			eventIndex++;
		}
		
		//*************************
		if (clip.x <= trackRightX) {
			// draw instance flag
			drawInstanceFlag(g2d, clip.x, 25, 53); //trackY, trackHeight);
			// draw event flag
			if (activeLLEvent >= 0 || activeHLEvent >= 0) {				
				boolean isHighLevelEvent = (activeHLEvent >=0 ? true : false);
				int activeEventIndex = (isHighLevelEvent ? activeHLEvent : activeLLEvent);
				int eventX;
				if(isHighLevelEvent){
					eventX = trackPadding + (activeEventIndex * (highLevelEventWidth + spaceBetweenMoves)) + highLevelEventWidth / 2;
//				else
//					eventX = trackPadding + (activeEventIndex * (elementWidth + spaceBetweenSystemEvents)) ;
					drawEventFlag(g2d, activeEventIndex, eventX, trackY_highLevel + trackHeight + 5, 20, 
						isHighLevelEvent, isHighLevelEvent ? startX_HL: startX_HL);
				}
			}
		}

	}
	protected int mapHighLevelEventIndex(int x, int y) {
		if ((y >= trackY_highLevel) && (y <= (trackY_highLevel + trackHeight))) {
			x -= trackPadding;
			x /= (highLevelEventWidth + spaceBetweenMoves);
			if (replayResult != null){
				if ((x >= 0) && (x < replayResult.getSeqAlignment().size())) {
					return x;
				} else {
					return -1;
				}
			}
		}
		return -1;
	}
	protected int mapLowLevelEventIndex(int x, int y) {
		
		if ((y >= trackY_lowLevel) && (y <= (trackY_lowLevel + trackHeight))) {
			// y-coordinate matches, remap x to index
			x -= trackPadding;
			x /= (elementWidth + spaceBetweenSystemEvents);
			if (replayResult != null){
				if ((x >= 0) && (x < numberOfActions)) {
					return x;
				} else {
					return -1;
				}
			} else {
				if ((x >= 0) && (x < instance.size())) {
					return x;
				} else {
					return -1;
				}
			}
		} 
		return -1;

	}
	protected void drawInstanceFlag(Graphics2D g2d, int x, int y, int height) {
		
		String hlSize = "";
		if (replayResult != null){
			hlSize = "Case ID: " ;//replayResult.getSeqAlignment().size() + " events";
		} 
		// calculate width
		g2d.setFont(g2d.getFont().deriveFont(11f));
		FontMetrics fm = g2d.getFontMetrics();
		int nameWidth = fm.stringWidth(traceLabel);
		int width =  nameWidth + 15 ;
		width = Math.max(width, trackPadding - 10);
		// draw flag shadow
		int shadowOffset = 4;
		int[] xSCoords = new int[] { x, x + width - elementTriOffset + shadowOffset, x + width + shadowOffset,
				x + width - elementTriOffset + shadowOffset, x };
		int[] ySCoords = new int[] { y + shadowOffset, y + shadowOffset, y + (height / 2) + shadowOffset,
				y + height + shadowOffset, y + height + shadowOffset };
		g2d.setColor(new Color(0, 0, 0, 100));
		g2d.fillPolygon(xSCoords, ySCoords, 5);
		// draw flag background
		g2d.setColor(colorBgInstanceflag);
		int[] xCoords = new int[] { x, x + width - elementTriOffset, x + width, x + width - elementTriOffset, x };
		int[] yCoords = new int[] { y, y, y + (height / 2), y + height, y + height };
		g2d.fillPolygon(xCoords, yCoords, 5);
		// draw string
		int fontHeight = fm.getHeight();
		int fontOffset = (height - fontHeight - fontHeight) / 3;
		g2d.setColor(new Color(220, 220, 220)); 
		g2d.drawString(traceLabel, x + 5, y + fontOffset + fontHeight - 1);
		g2d.setColor(new Color(200, 200, 200));
		
		g2d.drawString(hlSize, x + 5, y + height - fontOffset - fontHeight * 2);
		g2d.drawString(replayResult.getCaseID(), x + 5, y + height - fontOffset - 3);
	}

	protected void drawEventFlag(Graphics2D g2d, int index, int x, int y, int height, boolean isHighLevelEvent, int startX) throws IndexOutOfBoundsException {
		if (replayResult != null){
			String name = "";
			if(isHighLevelEvent){
				Move move = replayResult.getSeqAlignment().get(index);
				if(move.isLogMove())
					name = Utility.getAttributeValue(move.getEvent(), "concept:name");
				else
					name = move.getTransition().toString();
			}else
				name = replayResult.getCompositeMovesWithAction().get(index).getAction();
			
			// calculate width
			g2d.setFont(g2d.getFont().deriveFont(14f));
			FontMetrics fm = g2d.getFontMetrics();
			int width = fm.stringWidth(name) + 10;
			
			// draw background
			g2d.setColor(colorBgEventFlag);
//			g2d.fillRect(x, y, width, height);
			// set color
			Color color = Color.BLACK;			
//			if(isHighLevelEvent)
//				color = getColor(replayResult.getSeqAlignment().get(index));
//			else
//				color = getColor(replayResult.getCompositeMovesWithAction().get(index));			
			g2d.setColor(color);
			// draw anchor line
			//g2d.drawLine(x, y, x, y + height);
			// draw strings
			int fontHeight = fm.getHeight();
			int fontOffset = (height - fontHeight - fontHeight - fontHeight) / 4;
			if(isHighLevelEvent)
				g2d.drawString(name, x - 10, trackY_highLevel - fontOffset - fontHeight + 2);
			else
				g2d.drawString(name, x - 10, trackY_lowLevel - fontOffset - fontHeight + 2);//+ trackHeight + fontOffset + fontHeight + 4);
		} else {
			XExtendedEvent ate = new XExtendedEvent(instance.get(index));
			XEventClass eventClass = info.getEventClasses().getClassOf(instance.get(index));
			int occurrence = (eventClass != null ? eventClass.size() : 0);
			//		int occurrence = summary.getLogEvents().findLogEvent(ate.getElement(), ate.getType()).getOccurrenceCount();
			double frequency = (maxOccurrenceCount == 0 ? 0.0 : (double) occurrence / (double) maxOccurrenceCount);
	
			String ateName = (ate.getName() != null ? ate.getName() : "<no name>");
			String ateTransition = (ate.getTransition() != null ? ate.getTransition() : "<no transition>");
			String ateResource = (ate.getResource() != null ? ate.getResource() : "<no resource>");
			String name = index + ": " + ateName + " (" + ateTransition + ")";
			String originator = ateResource + "; freq: " + format.format(frequency);
			Date ts = ate.getTimestamp();
			String timestamp;
			if (ts != null) {
				timestamp = dateFormat.format(ate.getTimestamp());
			} else {
				timestamp = "<no timestamp>";
			}
			// calculate width
			g2d.setFont(g2d.getFont().deriveFont(9f));
			FontMetrics fm = g2d.getFontMetrics();
			int widthA = fm.stringWidth(name) + 10;
			int widthB = fm.stringWidth(originator) + 10;
			int widthC = fm.stringWidth(timestamp) + 10;
			int width = (widthA > widthB) ? widthA : widthB;
			if (widthC > width) {
				width = widthC;
			}
			// draw background
			g2d.setColor(colorBgEventFlag);
			g2d.fillRect(x, y, width, height);
			// set color
			Color color = AlignmentConstants.MOVESYNCCOLOR;
			g2d.setColor(color);
			// draw anchor line
			g2d.drawLine(x, y, x, y + height);
			// draw strings
			int fontHeight = fm.getHeight();
			int fontOffset = (height - fontHeight - fontHeight - fontHeight) / 4;
			g2d.drawString(name, x + 5, y + fontOffset + fontHeight - 2);
			g2d.drawString(originator, x + 5, y + fontOffset + fontHeight + fontOffset + fontHeight - 3);
			g2d.drawString(timestamp, x + 5, y + height - fontOffset - 2);
		} 
	}


	private Color getColor(Move move) {
		if(move.isLogMove())		 
				return AlignmentConstants.MOVELOGCOLOR;
		else if(move.isSilentModelMove())
				return AlignmentConstants.MOVEMODELINVICOLOR;
		else if(move.isModelMove())
				return colorModelMove;//Color(102,178,255);
		else if(move.isSyncMove())	 
				return colorModelMove;
		return Color.WHITE; // unknown
		
	}
	private Color getColor(CompositeMove compositeMove){
		if(compositeMove.hasAllowedAction())
			return AlignmentConstants.MOVESYNCCOLOR;
		else if(compositeMove.hasMissingAction())
			return AlignmentConstants.MOVESYNCCOLOR;
		else if(compositeMove.hasUnauthorizedAction())
			return AlignmentConstants.MOVELOGCOLOR;
		return Color.WHITE;
	}

	protected void drawEvent(Graphics2D g2d, int index, boolean active, int x, int y, int width, int height, boolean isHighLevel) {		
		Color color;
		if (replayResult != null){
			// set correct color for event
			color = getColorOfIndex(index, isHighLevel);
		} else {
			color = AlignmentConstants.MOVESYNCCOLOR;
		}
		
		if (active == false) {
			color = attenuateColor(color);
		}
		g2d.setColor(color);
		// draw triangularish shape
		int midPointBX = x + elementTriOffset;
		int midPointAX = x + width + elementTriOffset;
		int midPointY = y + (height / 2);
		int[] xCoords = new int[] { x, x + width, midPointAX, x + width, x, midPointBX };
		int[] yCoords = new int[] { y, y, midPointY, y + height, y + height, midPointY };
		int minus = 5;
		
		int[] xCoords2 = new int[] { x - minus, x + width - minus, midPointAX - minus, x + width - minus, x - minus, midPointBX - minus};
		int[] yCoords2 = new int[] { y + minus, y + minus, midPointY + minus, y + height + minus, y + height + minus, midPointY + minus};
		
		if (active == true) {
			for (int i = 0; i < xCoords.length; i++) {
				xCoords[i] -= 1;
				yCoords[i] -= 3;
			}
		}
		if(isHighLevel){
			if(replayResult != null && isHighLevel && replayResult.getSeqAlignment().get(index).isSyncMove()){
				if(active)
					g2d.setColor(AlignmentConstants.MOVELOGCOLOR);
				else
					g2d.setColor(attenuateColor(AlignmentConstants.MOVELOGCOLOR));
				g2d.fillPolygon(xCoords2, yCoords2, 6);
				g2d.setColor(Color.black);
				g2d.setStroke(boarder);
				g2d.drawPolygon(xCoords2, yCoords2, 6);
			}
			g2d.setColor(color);
			g2d.fillPolygon(xCoords, yCoords, 6);
			g2d.setColor(Color.black);
			g2d.setStroke(boarder);
			g2d.drawPolygon(xCoords, yCoords, 6);
			//draw links
			if(active || visualizationPanel.getLinkageComboBox().getSelectedIndex() == InterLevelReplayResultVisualizationPanel.SHOW_ALL_LINKS){
				for(int j = 0; j < replayResult.getCompositeMovesWithAction().size(); j++){
					if(replayResult.getCompositeMovesWithAction().get(j).getAlignmentMove() == null)
						continue;
					if(replayResult.getSeqAlignment().get(index).equals(replayResult.getCompositeMovesWithAction().get(j).getAlignmentMove())){
						Stroke stroke = null;
						if(replayResult.getCompositeMovesWithAction().get(j).getCRUDEntry() == null )
							stroke = dashed;
						else
							stroke = lineStroke;
						
						g2d.setStroke(stroke);
						if(active)
							g2d.setColor(Color.black);
						else
							g2d.setColor(Color.gray);
						g2d.drawLine( x + width/2, trackY_highLevel + trackHeight, 
								trackPadding + ((elementWidth + spaceBetweenSystemEvents) * j) + (elementWidth / 2), trackY_lowLevel);						
					}
				}
			}
		}else{
			CompositeMoveType cmType = replayResult.getCompositeMovesWithAction().get(index).getIDPStepType();
			if(cmType.equals(CompositeMoveType.MLOG_PROC_SYNC_DATA) || cmType.equals(CompositeMoveType.MMODEL_PROC_SYNC_DATA) ||
					cmType.equals(CompositeMoveType.SYNC_PROC_SYNC_DATA)){
				if(active)
					g2d.setColor(AlignmentConstants.MOVELOGCOLOR);
				else
					g2d.setColor(attenuateColor(AlignmentConstants.MOVELOGCOLOR));			
				int[] xCoords_diamond2 = new int[] { x - minus, x + width/2 - minus, x + width - minus, x + width/2 - minus};
				int[] yCoords_diamond2 = new int[] { midPointY , y , midPointY  , y + height  };
				g2d.fillPolygon(xCoords_diamond2, yCoords_diamond2, 4);
				g2d.setStroke(boarder);
				g2d.setColor(Color.black);
				g2d.drawPolygon(xCoords_diamond2, yCoords_diamond2, 4);
			}
			int[] xCoords_diamond = new int[] { x, x + width/2, x + width, x + width/2};
			int[] yCoords_diamond = new int[] { midPointY, y, midPointY, y + height};
			g2d.setColor(color);
			g2d.fillPolygon(xCoords_diamond, yCoords_diamond, 4);
			g2d.setStroke(boarder);
			g2d.setColor(Color.black);
			g2d.drawPolygon(xCoords_diamond, yCoords_diamond, 4);
			if(active){
				int labelOffset = 0; 
				g2d.setColor(Color.black);
				String name = replayResult.getCompositeMovesWithAction().get(index).getAction();			
				int fontHeight = g2d.getFontMetrics().getHeight();
				int fontOffset = (height - fontHeight - fontHeight - fontHeight) / 4;
				String arr[] = name.split(":");
				if(arr[0].length() < arr[1].length())
				{
					arr[0] = arr[0].toLowerCase();
					String tmp = arr[1];
					arr[1] = arr[0];
					arr[0] = tmp;
				}else
					arr[1] = arr[1].toLowerCase();
				for (int i = 0; i < 2; i++) {					
					if(arr[i].contains("(")){
						arr[i] = arr[i].replace(")", "");
						arr[i] = arr[i].split("\\(")[1];
					}
				}
				AttributedString as = new AttributedString(arr[0] + arr[1]);
				
				as.addAttribute(TextAttribute.SUPERSCRIPT, TextAttribute.SUPERSCRIPT_SUPER, arr[0].length(), arr[0].length() + 1);
//				name = arr[0] + ":" + arr[1];
				g2d.drawString(as.getIterator(), x + width/2 - 5, trackY_lowLevel - fontOffset - fontHeight + 34 + labelOffset + (eventNum % 2 ) * 10);
				eventNum++;
			}
		}
	}
	
	private Color getColorOfIndex(int index, boolean isHighLevel) {
		if(isHighLevel){
			return getColor(replayResult.getSeqAlignment().get(index));
		}
		return getColor(replayResult.getCompositeMovesWithAction().get(index));
	}

	protected Color attenuateColor(Color color) {
		int red = (int) (color.getRed() * 0.8);
		int green = (int) (color.getGreen() * 0.8);
		int blue = (int) (color.getBlue() * 0.8);
		return new Color(red, green, blue);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent
	 * )
	 */
	public void mouseDragged(MouseEvent evt) {
		mouseX = evt.getX();
		mouseY = evt.getY();
		repaint();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
	 */
	public void mouseMoved(MouseEvent evt) {
		mouseX = evt.getX();
		mouseY = evt.getY();
		repaint();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	public void mouseClicked(MouseEvent arg0) {
		// ignore
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
	 */
	public void mouseEntered(MouseEvent arg0) {
		mouseOver = true;
		repaint();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
	 */
	public void mouseExited(MouseEvent arg0) {
		mouseOver = false;
		repaint();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
	 */
	public void mousePressed(MouseEvent arg0) {
		// ignore
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	public void mouseReleased(MouseEvent arg0) {
		// ignore
	}
}
