/**
 * 
 */
package org.processmining.security.plugins.visualization.sequence;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.NumberFormat;
import java.util.HashMap;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;
import javax.swing.table.DefaultTableModel;

import org.processmining.framework.plugin.Progress;
import org.processmining.framework.util.ui.widgets.ProMComboBox;
import org.processmining.framework.util.ui.widgets.ProMSplitPane;
import org.processmining.framework.util.ui.widgets.ProMTextField;
import org.processmining.framework.util.ui.widgets.WidgetColors;
import org.processmining.plugins.petrinet.visualization.AlignmentConstants;
import org.processmining.security.plugins.Constants;
import org.processmining.security.plugins.replayer.InterLevelReplayResult;
import org.processmining.security.plugins.replayer.InterLevelSyncReplayResult;

import com.fluxicon.slickerbox.factory.SlickerFactory;
import com.fluxicon.slickerbox.ui.SlickerScrollBarUI;

import info.clearthought.layout.TableLayout;


public class InterLevelReplayResultVisualizationPanel extends JPanel implements ActionListener{
	private static final long serialVersionUID = -5897513989508944234L;

	/**
	 * Pointers to property variable
	 */
	protected static int RELIABLEMIN = 0;
	protected static int RELIABLEMAX = 1;
	protected static int MIN = 2;
	protected static int MAX = 3;

	// standard deviation is calculated based on http://mathcentral.uregina.ca/QQ/database/QQ.09.02/carlos1.html
	protected static int SVAL = 4;
	protected static int MVAL = 6;
	protected static int SVALRELIABLE = 5;
	protected static int MVALRELIABLE = 7;
	protected static int PERFECTCASERELIABLECOUNTER = 8;
	
	private Color bgColor = new Color(192,192,192); 
	protected int iconSize = 20, shiftX = 4, shiftY = 2, elementTriOffset = 4;
	
	// this value has to be stored because it is used by actionListener
	protected int numReliableCaseInvolved = 0; 
	
	// total calculated values
	protected Map<String, Double[]> calculations = new HashMap<String, Double[]>();
	protected ProMComboBox<String> interLevelLinkageComboBox;
	protected JPanel allAlignmentPanel;
	protected JButton pageButton;
	protected ProMTextField pageTextField;
	protected JLabel totalPage;
	protected final int numOfalignmentsPerPage = 1000;
	protected int selectedPage = 1;
	public final static int SHOW_ALL_LINKS = 0, HIDE_LINKS = 1;	
	protected Progress progress;
	protected InterLevelReplayResult replayResult;
	protected ScrollBar alignmentOanemWithScrollBar;
//	protected ProMPropertiesPanel logAlignmentPanel;
	protected final DefaultTableModel reliableCasesTModel = new DefaultTableModel() {
		private static final long serialVersionUID = -4303950078200984098L;

		@Override
		public boolean isCellEditable(int row, int column) {
			return false;
		}
	};

	
	public InterLevelReplayResultVisualizationPanel(InterLevelReplayResult replayResult, Progress progress) {
		this.replayResult = replayResult;
		this.progress = progress;
		TableLayout mainLayout = new TableLayout(new double[][] { { TableLayout.FILL }, { TableLayout.FILL } });
		setLayout(mainLayout);
		setBorder(BorderFactory.createEmptyBorder());
		if (progress != null) {
			progress.setMaximum(Math.min(numOfalignmentsPerPage, replayResult.size()));
		}
		pageTextField = new ProMTextField("1");
		pageButton = SlickerFactory.instance().createButton("Go");		
		add(createBottomPanel(), "0,0");
	}
	public class ScrollBar extends JScrollPane {
		private static final long serialVersionUID = 1L;

		public ScrollBar(JComponent component){
			super(component);
			setOpaque(true);
			setBackground(WidgetColors.PROPERTIES_BACKGROUND);
			getViewport().setOpaque(true);
			getViewport().setBackground(WidgetColors.PROPERTIES_BACKGROUND);
			setBorder(BorderFactory.createEmptyBorder());
			setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
			setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
			JScrollBar hBar = getHorizontalScrollBar();
			hBar.setUI(new SlickerScrollBarUI(hBar, new Color(0, 0, 0, 0), new Color(160, 160, 160),
					WidgetColors.COLOR_NON_FOCUS, 4, 12));
			hBar.setOpaque(true);
			hBar.setBackground(WidgetColors.PROPERTIES_BACKGROUND);

			
			JScrollBar vBar = getVerticalScrollBar();
			vBar.setUI(new SlickerScrollBarUI(vBar, new Color(0, 0, 0, 0), new Color(160, 160, 160),
					WidgetColors.COLOR_NON_FOCUS, 4, 12));
			vBar.setOpaque(true);
			vBar.setBackground(WidgetColors.PROPERTIES_BACKGROUND);
		}
	}
	private Component createBottomPanel() {
		// coloring scheme
//		Color bgColor = new Color(30, 30, 30);
		
		// clear maps
		calculations.clear();

		// add util
		SlickerFactory factory = SlickerFactory.instance();

		// for each case, create comparison panel

		final NumberFormat nf = NumberFormat.getInstance();
		nf.setMaximumFractionDigits(2);
		nf.setMinimumFractionDigits(2);

		final NumberFormat nfi = NumberFormat.getInstance();
		nfi.setMaximumFractionDigits(0);
		nfi.setMinimumFractionDigits(0);

//		logAlignmentPanel = new ProMPropertiesPanel("Inter Data Process View");
		allAlignmentPanel = new JPanel();
		allAlignmentPanel.setLayout(new BoxLayout(allAlignmentPanel, BoxLayout.Y_AXIS));
		
		String link [] = {"Show All Links", "Hide Links"};
		interLevelLinkageComboBox = new ProMComboBox<String>(link);
		interLevelLinkageComboBox.addActionListener(this);
		interLevelLinkageComboBox.setMaximumSize(new Dimension(270, 40));
		interLevelLinkageComboBox.setPreferredSize(new Dimension(200, interLevelLinkageComboBox.getPreferredSize().height));
		interLevelLinkageComboBox.setMinimumSize(new Dimension(200, interLevelLinkageComboBox.getPreferredSize().height));
		
		addInterLevelAlignments();
		int lineNumberRight = 0; 
		JPanel statisticPanel = new JPanel();
		
		statisticPanel.setBackground(bgColor);
		double[][] rightMainPanelSize = new double[][] {
				{ TableLayout.PREFERRED },
				{ TableLayout.PREFERRED, 30, TableLayout.PREFERRED, 30, TableLayout.PREFERRED, TableLayout.PREFERRED,
						60, TableLayout.PREFERRED } };

		statisticPanel.setLayout(new TableLayout(rightMainPanelSize));

		// add LEGEND
		statisticPanel.add(createLegendPanel(), "0, " + lineNumberRight++ + ", c, t");		
		JLabel linkLabel = SlickerFactory.instance().createLabel("Filter Inter-Level Links:");
		statisticPanel.add(linkLabel, "0, " + lineNumberRight++ + ", c, t");
		statisticPanel.add(interLevelLinkageComboBox, "0, " + lineNumberRight++ + ", c, t");	
		// Page
		statisticPanel.add(createPagePanel(), "0, " + lineNumberRight++ + ", 0, " + lineNumberRight++ + ", c, t");	
		
		// MAIN PANEL
		ProMSplitPane splitPanel = new ProMSplitPane(ProMSplitPane.HORIZONTAL_SPLIT);
		splitPanel.setBorder(BorderFactory.createEmptyBorder());
		splitPanel.setOneTouchExpandable(true);
		
		alignmentOanemWithScrollBar = new ScrollBar(allAlignmentPanel);
		splitPanel.setLeftComponent(alignmentOanemWithScrollBar);//logAlignmentPanel.add(alignmentOanemWithScrollBar));
		splitPanel.setRightComponent(statisticPanel);
		statisticPanel.setBackground(Color.white);
		splitPanel.setResizeWeight(1.0);	
		return splitPanel;
	}
	protected void addInterLevelAlignments(){
		for (int i = (selectedPage - 1) * numOfalignmentsPerPage; i < Math.min(selectedPage * numOfalignmentsPerPage, replayResult.size()); i++) {
			InterLevelSyncReplayResult res = replayResult.get(i);
			if (progress != null) {
				progress.inc();
			}

			// ALIGNMENT PANEL  
			InterLevelInstanceConformanceView alignmentPanel = new InterLevelInstanceConformanceView(this, "", res, 14, 20);
			
			// set scroll pane 
			JScrollPane hscrollPane = new JScrollPane(alignmentPanel);
			hscrollPane.setOpaque(true);
			hscrollPane.setBackground(bgColor);
			hscrollPane.getViewport().setOpaque(true);
			hscrollPane.getViewport().setBackground(bgColor);
			hscrollPane.setBorder(BorderFactory.createEmptyBorder());
			hscrollPane.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
			hscrollPane.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
			
			JScrollBar hBar = hscrollPane.getHorizontalScrollBar();
			hBar.setUI(new SlickerScrollBarUI(hBar, new Color(0, 0, 0, 0), new Color(160, 160, 160),
					WidgetColors.COLOR_NON_FOCUS, 4, 12));
			hBar.setOpaque(true);
			hBar.setBackground(WidgetColors.PROPERTIES_BACKGROUND);
			hBar = hscrollPane.getVerticalScrollBar();
			hBar.setUI(new SlickerScrollBarUI(hBar, new Color(0, 0, 0, 0), new Color(160, 160, 160),
					WidgetColors.COLOR_NON_FOCUS, 4, 12));
			hBar.setOpaque(true);
			hBar.setBackground(WidgetColors.PROPERTIES_BACKGROUND);
			allAlignmentPanel.add(hscrollPane);
//			logAlignmentPanel.addProperty(alignmentStatsPanel, hscrollPane);
//			logAlignmentPanel.addProperty("Case ID: " + res.getAgetCaseID(), hscrollPane);
		}
	}
	protected Component createPagePanel(){
		JPanel pagePanel = new JPanel();
		pagePanel.setMinimumSize(new Dimension(200, 400));
		pagePanel.setBorder(BorderFactory.createEmptyBorder());
		pagePanel.setBackground(Color.white);//new Color(192, 192, 192));
		TableLayout layout = new TableLayout(new double[][] { { 0.60, TableLayout.FILL }, {0.5, 0.5}});
		layout.setVGap(1);
		
		pagePanel.setLayout(layout);

		totalPage = SlickerFactory.instance().createLabel("Choose page number (total pages:" + (int)Math.ceil(((double)replayResult.size()) / ((double)numOfalignmentsPerPage)) + ")");
		pagePanel.add(totalPage, "0,0,1,0,c, c");
		pageTextField.addActionListener(this);
		pageTextField.setMaximumSize(new Dimension(150, 40));
		pageTextField.setPreferredSize(new Dimension(100, pageTextField.getPreferredSize().height));
		pageTextField.setMinimumSize(new Dimension(80, pageTextField.getPreferredSize().height));		
		
		pageButton.setPreferredSize(new Dimension(60, pageTextField.getPreferredSize().height));
		
		pageButton.setMinimumSize(new Dimension(40, pageTextField.getPreferredSize().height));
		pageButton.addActionListener(this);		
				
		pagePanel.add(pageTextField, "0,1,r,c");
		pagePanel.add(pageButton, "1,1,l,c");
		
		
		return pagePanel;
	}
	protected Component createLegendPanel() {
		SlickerFactory factory = SlickerFactory.instance();

		JPanel legendPanel = new JPanel();
		legendPanel.setBorder(BorderFactory.createEmptyBorder());
		legendPanel.setBackground(Color.white);//new Color(192, 192, 192));
		TableLayout layout = new TableLayout(new double[][] { { 0.10, TableLayout.FILL }, {} });
		layout.setVGap(1);
		
		legendPanel.setLayout(layout);

		layout.insertRow(0, 0.2);

		int row = 1;

		layout.insertRow(row, TableLayout.PREFERRED);
		JLabel legend = SlickerFactory.instance().createLabel("LEGEND");
		legendPanel.add(legend, "0,1,1,1,c, c");
		row++;

		layout.insertRow(row, 0.2);

		layout.insertRow(row, TableLayout.PREFERRED);
		JPanel greenPanel = new JPanel(){
			protected void paintComponent(Graphics g2d) {
				int x = shiftX, width = getWidth() - shiftX - 2 - elementTriOffset, y = 0, height = getHeight() - shiftY - 1;
				int midPointBX = x + elementTriOffset;
				int midPointAX = x + width + elementTriOffset;
				int midPointY = y + (height / 2);
				
				int[] xCoords = new int[] { x, x + width, midPointAX, x + width, x, midPointBX };
				int[] yCoords = new int[] { y, y, midPointY, y + height, y + height, midPointY };
				
				y+= shiftY;
				int[] xCoords2 = new int[] { x - shiftX, x - shiftX + width, midPointAX - shiftX, x + width - shiftX, x - shiftX, midPointBX - shiftX };
				int[] yCoords2 = new int[] { y, y, midPointY, y + height, y + height, midPointY };
				g2d.setColor(attenuateColor(AlignmentConstants.MOVELOGCOLOR));
				g2d.fillPolygon(xCoords2, yCoords2, 6);	
				g2d.setColor(Color.black);
				g2d.drawPolygon(xCoords2, yCoords2, 6);
				g2d.setColor(attenuateColor(InterLevelInstanceConformanceView.colorModelMove));
				g2d.fillPolygon(xCoords, yCoords, 6);
				
				g2d.setColor(Color.black);
				g2d.drawPolygon(xCoords, yCoords, 6);
			}
		};
		greenPanel.setPreferredSize(new Dimension(iconSize, iconSize));
		greenPanel.setSize(new Dimension(iconSize, iconSize));
		legendPanel.add(greenPanel, "0," + row + ",r, c");
		JLabel syncLbl = factory.createLabel("-Synchronous Process Move");
		legendPanel.add(syncLbl, "1," + row++ + ",l, c");

		layout.insertRow(row, TableLayout.PREFERRED);
		JPanel greyPanel = new JPanel(){
			protected void paintComponent(Graphics g2d) {
				int x = shiftX / 2, width = getWidth() - shiftX - elementTriOffset, y = 0, height = getWidth() - shiftY;
				int midPointBX = x + elementTriOffset;
				int midPointAX = x + width + elementTriOffset;
				int midPointY = y + (height / 2);
				
				int[] xCoords = new int[] { x, x + width, midPointAX, x + width, x, midPointBX };
				int[] yCoords = new int[] { y, y, midPointY, y + height, y + height, midPointY };
				
				g2d.setColor(attenuateColor(AlignmentConstants.MOVEMODELINVICOLOR));
				g2d.fillPolygon(xCoords, yCoords, 6);
			}
		};
		greyPanel.setPreferredSize(new Dimension(iconSize, iconSize));
		greyPanel.setSize(new Dimension(iconSize, iconSize));
		legendPanel.add(greyPanel, "0," + row + ",r, c");		
		
		JLabel moveInviLbl = factory.createLabel("-Unobservable Process Move on Model");
		legendPanel.add(moveInviLbl, "1," + row++ + ",l, c");

		layout.insertRow(row, TableLayout.PREFERRED);
		JPanel purplePanel = new JPanel(){
			protected void paintComponent(Graphics g2d) {
				int x = shiftX / 2, width = getWidth() - shiftX - elementTriOffset, y = 0, height = getWidth() - shiftY;
				int midPointBX = x + elementTriOffset;
				int midPointAX = x + width + elementTriOffset;
				int midPointY = y + (height / 2);
				
				int[] xCoords = new int[] { x, x + width, midPointAX, x + width, x, midPointBX };
				int[] yCoords = new int[] { y, y, midPointY, y + height, y + height, midPointY };
				
				g2d.setColor(attenuateColor(InterLevelInstanceConformanceView.colorModelMove));
				g2d.fillPolygon(xCoords, yCoords, 6);
				
				g2d.setColor(Color.black);
				g2d.drawPolygon(xCoords, yCoords, 6);
			}
		};
		purplePanel.setPreferredSize(new Dimension(iconSize, iconSize));
		purplePanel.setSize(new Dimension(iconSize, iconSize));
		legendPanel.add(purplePanel, "0," + row + ",r, c");
		
		JLabel moveRealLbl = factory.createLabel("-Process Move on Model");
		legendPanel.add(moveRealLbl, "1," + row++ + ",l, c");

		layout.insertRow(row, TableLayout.PREFERRED);
		JPanel yellowPanel = new JPanel(){
			protected void paintComponent(Graphics g2d) {
				int x = shiftX / 2, width = getWidth() - shiftX - elementTriOffset, y = 0, height = getWidth() - shiftY;
				int midPointBX = x + elementTriOffset;
				int midPointAX = x + width + elementTriOffset;
				int midPointY = y + (height / 2);
				
				int[] xCoords = new int[] { x, x + width, midPointAX, x + width, x, midPointBX };
				int[] yCoords = new int[] { y, y, midPointY, y + height, y + height, midPointY };
				
				g2d.setColor(attenuateColor(AlignmentConstants.MOVELOGCOLOR));
				g2d.fillPolygon(xCoords, yCoords, 6);
				g2d.setColor(Color.black);
				g2d.drawPolygon(xCoords, yCoords, 6);
			}
		};
		yellowPanel.setPreferredSize(new Dimension(iconSize, iconSize));
		yellowPanel.setSize(new Dimension(iconSize, iconSize));
		legendPanel.add(yellowPanel, "0," + row + ",r, c");
		JLabel moveLogLbl = factory.createLabel("-Process Move on Log");
		legendPanel.add(moveLogLbl, "1," + row++ + ",l, c");

		String labels[] = {"Synchronous Data Move", "Data Move on Model", "Data Move on Log"};
		for(int i = 0 ; i < 3; i++){
			row = addLegendPanel(layout, row, legendPanel, labels[i], Constants.SYSTEM_LEVEL_MOVE_COLOR[i]);
		}
		
		return legendPanel;
	}
	protected Color attenuateColor(Color color) {
		int red = (int) (color.getRed() * 0.8);
		int green = (int) (color.getGreen() * 0.8);
		int blue = (int) (color.getBlue() * 0.8);
		return new Color(red, green, blue);
	}
	private int addLegendPanel(TableLayout layout, int row, JPanel legendPanel, final String label, Color color){
		layout.insertRow(row, TableLayout.PREFERRED);
		JPanel greenPanel = new JPanel(){
			protected void paintComponent(Graphics g2d) {
				if(label.equalsIgnoreCase("Synchronous Data Move")){
					int x = shiftX, width = getWidth() - shiftX, y = 0, height = getWidth() - shiftX;
					int[] xCoords = new int[] { x, x + width / 2 , x + width, x + width / 2 };
					int[] yCoords = new int[] { y + height / 2, y, y + (height / 2), y + height};
					
					int[] xCoords2 = new int[] { x - shiftX, x - shiftX + width/2, x + width - shiftX, x - shiftX + width/2 };
					int[] yCoords2 = new int[] { y + height / 2, y, y + height / 2, y + height};
					g2d.setColor(attenuateColor(AlignmentConstants.MOVELOGCOLOR));
					g2d.fillPolygon(xCoords2, yCoords2, 4);
					g2d.setColor(Color.black);
					g2d.drawPolygon(xCoords2, yCoords2, 4);
					g2d.setColor(attenuateColor(AlignmentConstants.MOVESYNCCOLOR));
					g2d.fillPolygon(xCoords, yCoords, 4);
					g2d.setColor(Color.black);
					g2d.drawPolygon(xCoords, yCoords, 4);					
				}else if(label.equalsIgnoreCase("Data Move on Model")){
					int x = shiftX/2, width = getWidth() - shiftX, y = 0, height = getWidth() - shiftX;
					int[] xCoords = new int[] { x, x + width/2, x + width, x + width/2 };
					int[] yCoords = new int[] { y + height / 2, y, y + height / 2, y + height};
					
					g2d.setColor(attenuateColor(AlignmentConstants.MOVESYNCCOLOR));
					g2d.fillPolygon(xCoords, yCoords, 4);
					g2d.setColor(Color.black);
					g2d.drawPolygon(xCoords, yCoords, 4);
				}else if(label.equalsIgnoreCase("Data Move on Log")){
					int x = shiftX/2, width = getWidth() - shiftX, y = 0, height = getWidth() - shiftX;
					
					int[] xCoords2 = new int[] { x, x + width/2, x + width, x  + width/2 };
					int[] yCoords2 = new int[] { y + height / 2, y, y + height / 2, y + height};
					g2d.setColor(attenuateColor(AlignmentConstants.MOVELOGCOLOR));
					g2d.fillPolygon(xCoords2, yCoords2, 4);	
					g2d.setColor(Color.black);
					g2d.drawPolygon(xCoords2, yCoords2, 4);
				}
			}
	
		};
		greenPanel.setPreferredSize(new Dimension(iconSize, iconSize));
		greenPanel.setSize(new Dimension(iconSize, iconSize));
		
		legendPanel.add(greenPanel, "0," + row + ",r, c");
		JLabel syncLbl = SlickerFactory.instance().createLabel(label);
		legendPanel.add(syncLbl, "1," + row++ + ",l, c");
		
		
		return row;
	}

	public void actionPerformed(ActionEvent e) {
		if(e.getSource() instanceof JComboBox){
			JComboBox<String> cb =  (JComboBox<String>) e.getSource();
	        switch(cb.getSelectedIndex()){
	        	case 0:
	        		allAlignmentPanel.repaint();
	        		break;
	        	case 1:
	        		allAlignmentPanel.repaint();
	        		break;
	        }
		}else if(e.getSource() instanceof JButton){
			progress.setValue(0);
			selectedPage = Integer.parseInt(pageTextField.getText());
			allAlignmentPanel.removeAll();
			addInterLevelAlignments();
			alignmentOanemWithScrollBar.revalidate();
		}
	}
	public ProMComboBox<String> getLinkageComboBox(){
		return interLevelLinkageComboBox;
	}
}
